
#include "gamex-sound.hpp"

SoundX::SoundX()
{
	samples_per_sec = 0;
	bits_per_sample = 0;
	length = 0;
	num_bytes = 0;
	num_samples = 0;	
	data = NULL;
	dx_index = -1;
	dx_refresh = true;
}

SoundX::~SoundX()
{
	if(GameX.GetDS()) // if we haven't already released all of DirectSound
		if (data!=NULL) {delete [] data; data = NULL;}
}

bool SoundX::ReadName (File& wav, Buffer &code, char *match_name)
{
	char buf[5];
	wav.ReadC (4, buf);
	if (strncmp (buf, match_name, 4)==0)	return true;
	else									return false;
}

bool SoundX::ReadSize (File& wav, Buffer &code, unsigned int &size)
{
	XBYTE4 fourbytes;
	wav.Read4 (fourbytes); 
	size = code.ReadL (fourbytes);
	return true;
}

bool SoundX::LoadWav (char *filename)
{
	XBYTE2 twobytes;
	XBYTE4 fourbytes;
	XBYTE2 format;
	unsigned int size;
	char buf[100];
	int n;

	File wav;
	Buffer code;

	if (wav.Open (filename, FILE_READ | FILE_RANDOM) != FILE_EXIST_NO) {
		code.SetOrder (BUFFER_ORDER_LBF);
		
		if (ReadName(wav, code, "RIFF")==false) {InvalidWav(); return false;}
		if (ReadSize(wav, code, size)==false)	{InvalidWav(); return false;}
		if (ReadName(wav, code, "WAVE")==false)	{InvalidWav(); return false;}	
		if (ReadName(wav, code, "fmt ")==false)	{InvalidWav(); return false;}
		if (ReadSize(wav, code, size)==false)	{InvalidWav(); return false;}
		// Read WAV Format Header (THIS READS 16 BYTES)
			wav.Read2 (twobytes); format = code.ReadI (twobytes);				// WAV Format Tag (1 = PCM) (IGNORED)
			if (format!=1) {wav.Close(); return false;}
			wav.Read2 (twobytes); num_channels = code.ReadI (twobytes);				// WAV Number of Channels
			wav.Read4 (fourbytes); samples_per_sec = code.ReadL (fourbytes);	// WAV Samples per Second
			wav.Read4 (fourbytes);												// WAV Avergae Bytes per Second (IGNORED)
			wav.Read2 (twobytes);												// WAV Block Align (IGNORED)
			wav.Read2 (twobytes); bits_per_sample = code.ReadI (twobytes);		// WAV Number of Channels			
			wav.ReadC (size-16, buf);
		while(wav.EndOfFile()==0) {
			wav.ReadC (4, buf);
			if (strncmp (buf, "data", 4)==0) {
				if (ReadSize(wav, code, size)==false)	{InvalidWav(); return false;}
				// Read WAV Data
				num_bytes = size;								// Load number of bytes in data (from file)				
				num_samples = num_bytes / ((bits_per_sample/8) * num_channels);	// Calculate number of samples
				length = (float) num_samples / (float) samples_per_sec;		// Calculate length of sound (in seconds)
				if (data!=NULL) {delete [] data; data = NULL;}
				data = new char[num_bytes];
				if (bits_per_sample <= 8) {
					for (n = 0; n < num_bytes; n+=num_channels) 
						wav.ReadC (num_channels, data + n);
				} else if (bits_per_sample <=16) {
					for (n = 0; n < num_bytes; n+=(num_channels<<1)) 
						wav.ReadC (num_channels<<1, data + n);
				}
			} else if (strncmp (buf, "fact", 4)==0) {
				wav.Read4 (fourbytes);	
				wav.Read4 (fourbytes);			
			}
		}		
				
		dx_index = -1; // fixes bug that can make playing any sound fail if a different sound has been loaded already
		dx_refresh = true;
		wav.Close ();
	} else {
		char disp[500];
		sprintf (disp, ".WAV File Not Found: %s\n\nMake sure paths and names of WAV files are correct. If you started the game outside of Visual C++, make sure the executable is in the correct location to find sounds.", filename); 
		debug.Output(disp,"\n");
#ifdef _DEBUG
		MessageBox (GameX.GetWindow(), disp, "GameX Error", MB_OK);
#endif
		return false;
	}	
	return true;
}

void SoundX::InvalidWav (void)
{
	char disp[500];
	sprintf (disp, ".WAV Format Invalid: \n\nInvalid WAV Format. Not all types of WAV files are supported. Try saving your WAV file with a different program, or with 'extra wav information' turned off."); 
	debug.Output(disp,"\n");
#ifdef _DEBUG
	MessageBox (GameX.GetWindow(), disp, "GameX Error", MB_OK);
#endif
}

void SoundX::CutLength (float new_length)
{
	int new_samples, new_bytes;
	char* new_data;

	if (data!=NULL && length > new_length) {
		new_samples = (int) (samples_per_sec * new_length);
		new_bytes = (bits_per_sample/8) * num_channels * new_samples;	
		new_data = new char[new_bytes];
		memcpy (new_data, data, new_bytes);
		if (data!=NULL) {delete [] data; data = NULL;}
		data = new_data;
		if (new_data!=NULL) {delete [] new_data; new_data = NULL;}
		num_bytes = new_bytes;
		num_samples = new_samples;
		length = new_length;
		dx_refresh = true;
	}
}

bool SoundX::NeedUpdate (void)
{
	if (dx_index==-1 || dx_refresh==true) return true;
	return false;
}


MusicX::MusicX()
{
	segment = NULL;
	segment_state = NULL;
	registered = false;
	vol = 0;
	playstate = 0;
}

MusicX::~MusicX()
{
	if(GameX.GetDMP()) { // if we haven't already released all of DirectMusic
		Release();
	}
}

void MusicX::Release (void)
{
	if(segment != NULL) {
		segment->SetParam(GUID_Unload, -1, 0, 0,(void*)GameX.GetDMP());
		segment->Release(), segment = NULL;
	}
	if(segment_state != NULL) segment_state->Release(), segment_state = NULL;
}

// loads a MIDI file of a given filename into this MusicX object
bool MusicX::LoadMidi (char *filename)
{
	// this function loads a midi segment
	DMUS_OBJECTDESC obj_desc;

	Release();

	// get current working directory
	char dir [_MAX_PATH];
	WCHAR w_dir [_MAX_PATH];
	if(_getcwd(dir, _MAX_PATH) == NULL)
		return false;

	MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,dir,-1,w_dir,_MAX_PATH);

	// tell the loader were to look for files
	if (FAILED(GameX.GetDML()->SetSearchDirectory(GUID_DirectMusicAllTypes, w_dir, FALSE)))
		return false;

	// convert filename to wide string
	WCHAR w_filename[_MAX_PATH];
	MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,filename,-1,w_filename,_MAX_PATH);

	// setup object description
	ZeroMemory(&obj_desc, sizeof(obj_desc));
	obj_desc.dwSize = sizeof(obj_desc);
	obj_desc.guidClass = CLSID_DirectMusicSegment;
	wcscpy(obj_desc.wszFileName, w_filename);
	obj_desc.dwValidData = DMUS_OBJ_CLASS | DMUS_OBJ_FILENAME;

	if(FAILED(GameX.GetDML()->GetObject(&obj_desc,IID_IDirectMusicSegment, (void**) &segment))
	|| FAILED(segment->SetParam(GUID_StandardMIDIFile,-1, 0, 0, (void*)GameX.GetDMP()))
	|| FAILED(segment->SetParam(GUID_Download, -1, 0, 0, (void*)GameX.GetDMP()))) {
		Release();
		return false;
	}

	segment->GetParam(GUID_TempoParam, 0xFFFF, 0, 0, NULL, &original_tempo);

	if(!registered) {
		GameX.RegisterMusicX(this);
		registered = true;
	}

	return true;
}


// sends a curve message (to alter volume or pitch over time)
void MusicX::Curve(long time, int target, int type, int shape, int start)
{
	DMUS_CURVE_PMSG *pCurveMsg;
	HRESULT hr;

	if (NULL == GameX.GetDMP()) return;
	hr = GameX.GetDMP()->AllocPMsg(sizeof(DMUS_CURVE_PMSG), (DMUS_PMSG**) &pCurveMsg);
	if (SUCCEEDED(hr))
	{
		ZeroMemory(pCurveMsg, sizeof(DMUS_CURVE_PMSG));
		pCurveMsg->dwSize = sizeof(DMUS_CURVE_PMSG);
		pCurveMsg->rtTime = 0;
		pCurveMsg->dwFlags = DMUS_PMSGF_REFTIME | DMUS_PMSGF_LOCKTOREFTIME;
		pCurveMsg->dwPChannel = DMUS_PCHANNEL_BROADCAST_PERFORMANCE;
		pCurveMsg->dwType = DMUS_PMSGT_CURVE;
		pCurveMsg->dwGroupID = 0xFFFFFFF;
		pCurveMsg->mtDuration = time;
		pCurveMsg->nEndValue = target;
		pCurveMsg->bCurveShape = shape;
		pCurveMsg->bCCData = 7;
		if(start == -1) {
			pCurveMsg->bFlags = DMUS_CURVE_START_FROM_CURRENT;
		} else {
			pCurveMsg->bFlags = 0;
			pCurveMsg->nStartValue = start;
		}
		pCurveMsg->bType = type ;
		pCurveMsg->wMergeIndex = merge_id++; // curves attempt to merge unless they have identical merge indexes
		hr = GameX.GetDMP()->SendPMsg((DMUS_PMSG*) pCurveMsg);
	}
}


